implementation module rc
import StdEnv, utils

:: *Buffer :== ( *{#Char} , !Int)

writeBytes [] buf = buf
writeBytes [b:bs] (buffer, pos) = writeBytes bs ({buffer & [pos] = b}, pos+1)

bufferContents :: *Buffer -> {#Char}
bufferContents (buf,pos) 
  | pos > 0 = { buf.[i] \\ i <- [0..pos-1] }
            = ""

newBuffer :: Int -> *Buffer
newBuffer sz = (createArray sz '-', 0)

readByte :: *Buffer -> (!Char, *Buffer)
readByte (buf, pos) = (ch, (buf1, pos+1)) where
  (ch, buf1) = buf![pos]

:: *Encoder = { buf :: !*Buffer 
              , low :: !Int 
              , range :: !Int 
              , ffnum :: !Int 
              , cache :: !Int 
              , carry :: !Int 
              }

:: *Decoder = { buf_d :: !*Buffer 
              , range_d :: !Int 
              , val :: !Int 
              }

      
TOP :== 0x10000
Thres :== 255*TOP
            
newEncoder :: Int -> Encoder
newEncoder sz = { buf = newBuffer sz, low = 0, 
  range = 0xFFFFFF,  ffnum = 0, cache = 0, carry = 0 }
    
encode :: !*Encoder !.Int !.Int !.Int -> *Encoder    
encode rc cumFreq freq totFreq = renorm rc1 where
  quant = rc.range / totFreq
  low1 = (rc.low + cumFreq * quant) bitand 0xFFFFFF
  rc1  = { rc & low = low1
         , range = quant * freq
         , carry = if (low1 < rc.low) (rc.carry+1) rc.carry
         }
  
renorm rc 
  | rc.range < TOP = renorm (shiftLow rc)
                   = rc

shiftLow rc = if (rc.low < Thres || rc.carry > 0) rc1 rc2 where
  range1 = rc.range << 8  
  low1 = (rc.low bitand 0xFFFF) << 8
  rc1 = { rc & range = range1, low = low1, ffnum = 0, cache = rc.low >> 16,
          buf = buf1, carry = 0 }  
  buf1 = writeBytes [toChar (rc.cache + rc.carry) : repeatn rc.ffnum (toChar (rc.carry-1))] rc.buf
  rc2 = { rc & low = low1, range = range1, ffnum = rc.ffnum+1 }  
  
finishEncode :: *Encoder -> {#Char}  
finishEncode rc = bufferContents rc1.buf where
  rc1 = iter 4 shiftLow rc  

newDecoder :: {#Char} -> *Decoder
newDecoder str = iter 4 loadByte rc where
  rc = { buf_d = (buf,0),  range_d = 0xFFFFFF, val = 0 }
  buf = {c \\ c <-: str}
  
loadByte rc 
  # (c, buf1) = readByte rc.buf_d
  = { rc & buf_d = buf1, val = ((rc.val bitand 0xFFFF) << 8) + toInt c }

getFreq :: !*Decoder !.Int -> *(Int,*Decoder)
getFreq rc totFreq 
  #  rc = { rc & range_d = rc.range_d / totFreq }
  #! freq = rc.val / rc.range_d
  = (freq, rc)

decode :: !*Decoder !.Int !.Int -> *Decoder
decode rc cumFreq freq = load rc1 where
  rc1 = { rc & val = rc.val - cumFreq * rc.range_d, range_d = rc.range_d * freq }
  
load rc
  | rc.range_d < TOP = load (loadByte {rc & range_d = rc.range_d << 8})
                     = rc
    
