module zorc
import StdEnv, rc, utils, ArgEnv

countSymbols text = foldArray f a0 text where
  a0 = createArray 256 0
  f a c 
    # i = toInt c
    # (old, a) = a![i]
    = {a & [i] = old + 1}

normstat cntab = if (total < 0x10000) cntab scaled where
  total = foldArray (+) 0 cntab
  scaled = {f x  \\ x <-: cntab}
  f x = if (x==0) 0 (toInt (toReal x * k + 0.5) + 1)
  k = toReal 0xFD00 / toReal total 

calcstats text 
  # cntab = countSymbols text |> normstat
    totFreq = foldArray (+) 0 cntab
    cumFreqs = calcCF cntab
  = (cntab, cumFreqs, totFreq) 
where  
  calcCF :: .{#Int} -> .{#Int}
  calcCF tab = { cf i \\ i <- [0..255]} where
    cf i = sum [tab.[j] \\ j <- [0..i-1]]

writeStats f stat = foldArray writecnt f stat where
  writecnt f n = fwritec c1 f |> fwritec c2 where
    c1 = n bitand 255 |> toChar
    c2 = n >> 8 |> toChar

readStats :: *File -> ({#Int}, *File)
readStats f = rsloop 0 f tab0 where
  tab0 :: *{#Int}
  tab0 = createArray 256 0
  rsloop i f tab
    | i >= 256 = ({x \\ x<-:tab}, f)
    # (ok, c1, f) = freadc f
      (ok, c2, f) = freadc f
      tab = {tab & [i] = toInt c2 << 8 + toInt c1}
    = rsloop (i+1) f tab

compress text = (cntab, compressed) where 
  (cntab, cumFreqs, totFreq) = calcstats text
  compressed = finishEncode rc
  rc0 = newEncoder (size text)
  rc = foldArray encodeSymb rc0 text  
  encodeSymb rc c  = encode rc cumFreq freq totFreq where  
    i = toInt c
    freq = cntab.[i]
    cumFreq = cumFreqs.[i]
    
findSymb tab val = fsloop 0 0 where
  fsloop i cumfr 
    #!nf = cumfr + tab.[i]
    | nf > val = (i, cumfr, tab.[i])
    = fsloop (i+1) nf

    
decompress cntab compressed len = decompressed where
  rc0 = newDecoder compressed
  totFreq = foldArray (+) 0 cntab
  buf0 :: *{#Char}
  buf0 = createArray len '-'
  decompressed = decloop 0 rc0 buf0 
  decloop pos rc buf 
    | pos >= len = buf
    #!(val, rc) = getFreq rc totFreq
    #!(symb, cumfr, fr) = findSymb cntab val
    #!rc = decode rc cumfr fr
    #!buf = {buf & [pos] = toChar symb}
    = decloop (pos+1) rc buf 

fileSize :: *File -> *(!Int, *File)  
fileSize f
  # (ok, f) = fseek f 0 FSeekEnd
  # (pos, f) = fposition f 
  # (ok, f) = fseek f 0 FSeekSet
  = (pos, f)


Start :: *World -> *World
Start world
  #!cmdline = getCommandLine 
  | size cmdline < 4 = abort "zorc {c|d} infile outfile"
  # (ok, f, world) = fopen cmdline.[2] FReadData world
  | ok==False = abort "cannot open file"
  # (ok, fo, world) = fopen cmdline.[3] FWriteData world
  | ok==False = abort "cannot open out file"
  | cmdline.[1] == "c"   
     # (sz, f) = fileSize f
       (text, f) = freads f sz
       (_, world) = fclose f world
       (stats, compressed) = compress text
       fo = fwritei (size text) fo
       fo = writeStats fo stats 
       fo = fwrites compressed fo
       (ok, world) = fclose fo world
     = world
  | cmdline.[1] == "d"     
     # (ok,len,f) = freadi f
       (stats, f) = readStats f 
       (compressed, f) = freads f len
       (_, world) = fclose f world
       decompressed = decompress stats compressed len
       fo = fwrites decompressed fo
       (ok, world) = fclose fo world
     = world
